from ..lib.constants import DEFAULT_INPUT_HANDLE_VALUE, DEFAULT_OUTPUT_HANDLE_VALUE
from .find import find_by_attr

def get_default_input_val(inputs_with_handle_name: dict, default_val = None):
  if DEFAULT_INPUT_HANDLE_VALUE in inputs_with_handle_name:
    return inputs_with_handle_name[DEFAULT_INPUT_HANDLE_VALUE] or default_val
  for hanldle in inputs_with_handle_name:
    return inputs_with_handle_name[hanldle] or default_val
  return default_val

def get_default_handle(handles, handle_type = 'inputs'):
  if not handles:
    return None
  default_handle_name = DEFAULT_INPUT_HANDLE_VALUE if handle_type == 'inputs' else DEFAULT_OUTPUT_HANDLE_VALUE
  handle_list = handles[handle_type] or []
  default_handle = find_by_attr(handle_list, 'handle', default_handle_name)
  if default_handle:
    return default_handle
  elif len(handle_list) > 0:
    return handle_list[0]
  return None



from .utils import ComponentABC

class EventListener(ComponentABC):
    def __init__(self, agent: object):
        self.agent = agent
        self.listeners = {}

    def add(self, event:str, listener: callable):
        if event not in self.listeners:
            self.listeners[event] = []
        self.listeners[event].append(listener)
        return self.agent

    def on_delta(self, listener: callable):
        self.add("response:delta", listener)
        return self.agent

    def on_done(self, listener: callable):
        self.add("response:done", listener)
        return self.agent

    def on_finally(self, listener: callable):
        self.add("response:finally", listener)
        return self.agent

    def _call(self, event:str, data: any):
        if event in self.listeners:
            for listener in self.listeners[event]:
                listener(data)
        if event == "response:finally":
            self.listeners = {}

    def export(self):
        return {
            "prefix": None,
            "suffix": self._call,
            "alias": {
                "add_event_listener": { "func": self.add },
                "on_delta": { "func": self.on_delta },
                "on_done": { "func": self.on_done },
                "on_finally": { "func": self.on_finally },
            },
        }

def export():
    return ("EventListener", EventListener)


import inspect
import uuid
import copy
from .utils.find import has_target_by_attr
from .lib.constants import DEFAULT_INPUT_HANDLE, DEFAULT_OUTPUT_HANDLE, EXECUTOR_TYPE_NORMAL

SPECIAL_CHUNK_TYPES = ['Start']
class SchemaChunk:
    """
    Workflow Schema 中的单个 chunk 结构，用于辅助提供操作的 API，实际运行时，提供 chunk 描述参数交给 MainExecutor 执行
    """

    def __init__(self, workflow_schema = None, type = EXECUTOR_TYPE_NORMAL, executor: callable = None, **chunk_desc):
        """
        可选参数 handles(连接点，{'inputs': [], 'outputs': []}) 字段，如没有 handles 字段，则会自动追加上默认设置、interactions、title，必选参数 executor，或 type 为 'Start'
        """
        # 校验必填字段(要么 type 为特殊类型，否则必须包含 executor)
        if not type or (type not in SPECIAL_CHUNK_TYPES):
            if not executor:
                raise ValueError("Missing required key: 'executor'")
        
        handles = chunk_desc.get('handles') or {
            'inputs': [DEFAULT_INPUT_HANDLE.copy()],
            'outputs': [DEFAULT_OUTPUT_HANDLE.copy()]
        }
        if 'inputs' not in handles:
            handles['inputs'] = [DEFAULT_INPUT_HANDLE.copy()]
        if 'outputs' not in handles:
            handles['outputs'] = [DEFAULT_OUTPUT_HANDLE.copy()]

        self.chunk = {
            'id': chunk_desc.get('id', str(uuid.uuid4())),
            'title': chunk_desc.get('title'),
            'interactions': chunk_desc.get('interactions') or {}, # 交互配置
            'type': type or EXECUTOR_TYPE_NORMAL,
            'executor': executor,
            'handles': handles,
            # 连接条件（默认无条件连通）
            'connect_condition': chunk_desc.get('connect_condition') or None,
            # 当前激活的待连接的 handle 名
            'active_handle': chunk_desc.get('active_handle') or None
        }

        self.workflow_schema = workflow_schema

        # 解析出默认的输入连接点(有 input 直接用 input 做默认，否则取第一个手柄为默认连接点)
        inputs = self.chunk.get('handles').get('inputs', [])
        self.default_input_handle = None
        if has_target_by_attr(inputs, 'handle', DEFAULT_INPUT_HANDLE['handle']):
            self.default_input_handle = DEFAULT_INPUT_HANDLE['handle']
        elif len(inputs) > 0:
            self.default_input_handle = inputs[0]['handle']

        # 解析出默认的输出连接点(有 output 直接用 output 做默认，否则取第一个手柄为默认连接点)
        outputs = self.chunk.get('handles').get('outputs', [])
        self.default_output_handle = None
        if has_target_by_attr(outputs, 'handle', DEFAULT_OUTPUT_HANDLE['handle']):
            self.default_output_handle = DEFAULT_OUTPUT_HANDLE['handle']
        elif len(outputs) > 0:
            self.default_output_handle = outputs[0]['handle']
    
    def handle(self, name: str):
        """
        Chunk 的连接点
        """
        if not self.has_handle(name) and not self.has_handle(name, False):
            raise ValueError(f"The handle '{name}' for chunk '{self.chunk.get('title', self.chunk['id'])}' does not exist.")
        shadow_chunk = SchemaChunk(
            workflow_schema=self.workflow_schema,
            **self.get_raw_schema()
        )
        shadow_chunk.set_active_handle(name)
        return shadow_chunk
    
    def if_condition(self, condition: callable = None):
        """
        按条件连接
        """
        shadow_chunk = SchemaChunk(
            workflow_schema=self.workflow_schema,
            **self.get_raw_schema()
        )
        shadow_chunk.set_connect_condition(condition)
        return shadow_chunk

    def else_condition(self):
        """
        按当前条件的反条件连接
        """
        current_condition = self.chunk.get('connect_condition')
        if not current_condition:
            current_condition = lambda values: True
        else_condition_func = lambda values: not current_condition(values)
        shadow_chunk = SchemaChunk(
            workflow_schema=self.workflow_schema,
            **self.get_raw_schema()
        )

        shadow_chunk.set_connect_condition(else_condition_func)
        return shadow_chunk

    def connect_to(self, chunk):
        """
        连接到指定个 Chunk 节点，支持传入 create_chunk 创建好的实例，也支持传入 chunk dict
        """
        # 如果传入的是一个 dict 类型，自动进行实例化再连接
        target_chunk = self.workflow_schema.create_chunk(chunk) if isinstance(chunk, dict) else chunk
        if not isinstance(target_chunk, SchemaChunk):
            raise ValueError(f"The 'chunk' parameter to 'connect_to' is invalid.")

        # 检测 chunk 是否在当前 workflow 中有注册
        if not self.workflow_schema.get_chunk(target_chunk.chunk['id']):
            raise ValueError(f"The two chunk are not mounted under the same 'workflow'.")

        # 解析出连接手柄，并判断合法性
        expect_source_handle = self.chunk.get('active_handle') or self.default_output_handle
        if not self.has_handle(expect_source_handle, False):
            raise ValueError(f"The source chunk's connection handle '{expect_source_handle}' does not exist.")

        expect_target_handle = target_chunk.chunk.get('active_handle') or target_chunk.default_input_handle
        if not target_chunk.has_handle(expect_target_handle):
            raise ValueError(f"The target chunk's connection handle '{expect_target_handle}' does not exist.")
        
        # 连接两 chunk
        self.workflow_schema.connect_chunk(
            self.chunk['id'],
            target_chunk.chunk['id'],
            expect_source_handle,
            expect_target_handle,
            self.chunk.get('connect_condition') or None
        )
        return self
    
    def get_raw_schema(self):
        return copy.deepcopy(self.chunk)
    
    def set_active_handle(self, name: str = None):
        """
        设置激活的连接点，连接时会默认连接到该点上
        """
        if name is not None:
            if not self.has_handle(name) and not self.has_handle(name, False):
                raise ValueError(
                    f"The handle '{name}' for chunk '{self.chunk.get('title', self.chunk['id'])}' does not exist.")
        self.chunk['active_handle'] = name
    
    def set_connect_condition(self, condition: callable = None):
        """
        设置连接的条件
        """
        self.chunk['connect_condition'] = condition
    
    def has_handle(self, name: str, from_inputs = True) -> bool:
        """
        判断连接手柄是否存在
        """
        search_range = self.chunk.get('handles').get('inputs' if from_inputs else 'outputs', [])
        return has_target_by_attr(search_range, 'handle', name)


from .utils.verify import validate_dict
from .utils.find import has_target_by_attr, find_by_attr
from .Chunk import SchemaChunk
from .lib.constants import DEFAULT_INPUT_HANDLE_VALUE, DEFAULT_OUTPUT_HANDLE_VALUE, DEFAULT_OUTPUT_HANDLE, DEFAULT_INPUT_HANDLE, EXECUTOR_TYPE_NORMAL

class Schema:
    """
    Workflow 的描述
    """

    def __init__(self, schema_data = { 'chunks': [], 'edges': [] }):
        self.chunks = []
        self.edges = []
        # 依次调用添加方法添加
        (
            self
                .append_raw_chunk_list(schema_data.get('chunks', []))
                .connect_with_edges(schema_data.get('edges', []))
        )
    
    def create_chunk(self, type = EXECUTOR_TYPE_NORMAL, executor: callable = None, **chunk_desc) -> SchemaChunk:
        """
        根据 chunk 的描述，创建 chunk 实例
        """
        chunk_inst = SchemaChunk(
            workflow_schema=self,
            type=type,
            executor=executor,
            **chunk_desc
        )
        self.append_raw_chunk(chunk_inst.get_raw_schema())
        return chunk_inst

    def append_raw_chunk(self, chunk: dict):
        """
        添加节点，必须包含 id, type, handles(连接点，{'inputs': [], 'outputs': []}) 字段，如没有 handles 字段，则会自动追加上默认设置
        """
        # 校验必填字段
        verified_res = validate_dict(chunk, ['id', 'type'])
        if verified_res['status'] == False:
            raise ValueError(f"Missing required key: '{verified_res['key']}'")

        chunk_copy = chunk.copy()
        # 如果没有配置 handles，则自动追加上
        if 'handles' not in chunk_copy:
            chunk_copy['handles'] = {
                'inputs': [DEFAULT_INPUT_HANDLE.copy()],
                'outputs': [DEFAULT_OUTPUT_HANDLE.copy()]
            }
        else:
            if 'inputs' not in chunk_copy['handles']:
                chunk_copy['handles']['inputs'] = [DEFAULT_INPUT_HANDLE.copy()]
            if 'outputs' not in chunk_copy['handles']:
                chunk_copy['handles']['outputs'] = [
                    DEFAULT_OUTPUT_HANDLE.copy()]
        
        # 检查是否有重名的 chunk
        if has_target_by_attr(self.chunks, 'id', chunk_copy['id']):
            raise ValueError(
                f"The Chunk with the existing id '{chunk_copy['id']}' already exists.")
        self.chunks.append(chunk)
    
    def append_raw_chunk_list(self, chunks: list):
        for chunk in chunks:
            self.append_raw_chunk(chunk)
        return self

    def del_chunk(self, chunk_id):
        """
        删除节点
        """
        self.chunks = [
            chunk for chunk in self.chunks if chunk.get("id") != chunk_id]
        # 清理相关的边(起点/终点为当前节点的边)
        self.edges = [
            edge for edge in self.edges if edge.get("source") != chunk_id and edge.get("target") != chunk_id]
    
    def get_chunk(self, chunk_id):
        return find_by_attr(self.chunks, 'id', chunk_id)

    def connect_chunk(self, source_chunk_id, target_chunk_id, source_handle=DEFAULT_OUTPUT_HANDLE_VALUE, target_handle=DEFAULT_INPUT_HANDLE_VALUE, condition: callable = None):
        """
        连接两个节点，分别输入：源节点id、目标节点id、源节点输出句柄（可选，默认为源节点的输出点）、目标节点的输入句柄（可选，默认为目标节点的输入点）
        """
        if source_chunk_id == target_chunk_id:
            raise ValueError(
                f"The starting point and the endpoint have the same value: '{target_chunk_id}'")
        source_chunk = None
        target_chunk = None
        for chunk in self.chunks:
            if chunk['id'] == source_chunk_id:
                source_chunk = chunk
            elif chunk['id'] == target_chunk_id:
                target_chunk = chunk

        # 判断节点是否存在
        if source_chunk is None or target_chunk is None:
            raise ValueError('Cannot find corresponding start or end node')

        # 判断句柄是否存在
        source_output_handles = source_chunk.get(
            'handles', {'inputs': [], 'outputs': []}).get('outputs', [])
        if has_target_by_attr(source_output_handles, 'handle', source_handle) == False:
            raise ValueError(
                f'Source node "{source_chunk.get("title", source_chunk_id)}({source_chunk_id})" lacks output endpoint: {source_handle}')

        target_output_handles = target_chunk.get(
            'handles', {'inputs': [], 'outputs': []}).get('inputs', [])
        if has_target_by_attr(target_output_handles, 'handle', target_handle) == False:
            raise ValueError(
                f'Target node "{target_chunk.get("title", target_chunk_id)}({target_chunk_id})" lacks input endpoint: {target_handle}')

        self.edges.append({
            'source': source_chunk_id,
            'target': target_chunk_id,
            'source_handle': source_handle or DEFAULT_OUTPUT_HANDLE_VALUE,
            'target_handle': target_handle or DEFAULT_INPUT_HANDLE_VALUE,
            'condition': condition # 连通的开关条件，默认直连
        })
    
    def get_edge(self, edge_id):
        return find_by_attr(self.edges, 'id', edge_id)
    
    def connect_with_edges(self, edges: list):
        """
        批量添加连接，要求输入为 {"source": str, "target": str, "source_handle"?: str, "target_handle"?: str} 形式构成的数组
        """
        for edge in edges:
            self.connect_chunk(
                edge.get('source'),
                edge.get('target'),
                edge.get('source_handle', DEFAULT_OUTPUT_HANDLE_VALUE),
                edge.get('target_handle', DEFAULT_INPUT_HANDLE_VALUE),
                edge.get('condition') or None
            )
        return self

    def del_connection(self, source_chunk_id, target_chunk_id, source_handle=DEFAULT_OUTPUT_HANDLE_VALUE, target_handle=DEFAULT_INPUT_HANDLE_VALUE):
        """
        取消连线，分别输入：源节点id、目标节点id、源节点输出句柄（可选，默认为源节点的输出点）、目标节点的输入句柄（可选，默认为目标节点的输入点）
        """
        edges = []
        for edge in self.edges:
            if not (
                edge.get("source") == source_chunk_id and
                edge.get("target") == target_chunk_id and
                (source_handle is None or edge.get('source_handle') == source_handle) and
                (target_handle is None or edge.get('target_handle') == target_handle)
            ):
                edges.append(edge)
        # 有变更时赋值
        if len(edges) != len(self.edges):
            self.edges = edges



# 入口类型
EXECUTOR_TYPE_START = 'Start'
# 结束类型
EXECUTOR_TYPE_END = 'End'
# 判断类型
EXECUTOR_TYPE_JUDGE = 'Judge'
# 常规类型
EXECUTOR_TYPE_NORMAL = 'Normal'
# 未知类型
EXECUTOR_TYPE_UNKNOW = 'Unknow'

# 默认的输入 handle
DEFAULT_INPUT_HANDLE_VALUE = 'input'
DEFAULT_INPUT_HANDLE = {'handle': DEFAULT_INPUT_HANDLE_VALUE, 'title': 'input'}
# 默认的输出 handle
DEFAULT_OUTPUT_HANDLE_VALUE = 'output'
DEFAULT_OUTPUT_HANDLE = {'handle': DEFAULT_OUTPUT_HANDLE_VALUE, 'title': 'output'}



from .Request import Request
from .Agent import AgentFactory
from .Facility import FacilityManager
from .WebSocket import WebSocketServer, WebSocketClient
from .Workflow import Workflow, Schema as WorkflowSchema
from ._global import global_plugin_manager, global_settings, global_storage, global_tool_manager, global_websocket_server
from .utils import *

def create_agent(*args, **kwargs):
    return AgentFactory().create_agent(*args, **kwargs)

facility = FacilityManager()

set_settings = global_settings.set


import json
import asyncio
import threading
import aiohttp
import importlib.metadata

async def check_version_async():
    current_version = importlib.metadata.version("Agently")
    async with aiohttp.ClientSession() as session:
        async with session.get(
            f"http://api.agently.tech/meta/api/latest?version={ current_version }",
            headers = { "Content-Type": "application/json" },
        ) as response:
            try:
                async for chunk in response.content.iter_chunks():
                    result = chunk[0].decode("utf-8")
                    result = json.loads(result)
                    stable_version = result["data"]["version"] if "version" in result["data"] else None
                    latest_version = result["data"]["latest"] if "latest" in result["data"] else None
                    tips_type = result["data"]["type"]
                    tips_content = result["data"]["tips"]
                    print("[Agently Version Check]")
                    print("This check only works 1 time each day in debug model.")
                    print(f"Current Installed Version: { current_version }")
                    if stable_version:
                        print(f"Stable Version: { stable_version }")
                    if latest_version:
                        print(f"Latest Version: { latest_version }")
                    if tips_type != "" and tips_content != "":
                        print(f"[{ tips_type.upper() }]{ tips_content }")
            except Exception as e:
                print(e)

def run_check_version():
    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)
    loop.run_until_complete(check_version_async())
    loop.close()

def check_version(global_storage, today):
    thread = threading.Thread(target = run_check_version)
    thread.start()
    #thread.join()
    global_storage.set("agently", "check_version_record", today)


from .utils import ToolABC
import time
import httpx
from duckduckgo_search import DDGS
from newspaper import Article as NsArticle, Config as NsConfig
#from bs4 import BeautifulSoup

class Web(ToolABC):
    def __init__(self, tool_manager: object):
        self.tool_manager = tool_manager

    def search(self, keywords: str, *, options: dict={}, proxy: str=None, type: int=1, no_sleep: bool=False, timelimit: str=None):
        if timelimit:
            options.update({ "timelimit": timelimit })
        results = []
        '''
        results = {
            "origin": [],
            "for_agent": [],
        }
        '''
        search_kwargs = {}
        if proxy:
            search_kwargs["proxies"] = proxy
        if "max_results" not in options:
            options.update({ "max_results": 5 })
        try:
            with DDGS(**search_kwargs) as ddgs:
                if type == 1:
                    for index, result in enumerate(ddgs.text(
                        str(keywords),
                        **options
                    )):
                        results.append({
                            "title": result["title"],
                            "brief": result["body"],
                            "url": result["href"],
                        })
                        '''
                        results["origin"].append(result)
                        results["for_agent"].append({
                            "index": index,
                            "title": result["title"],
                            "body": result["body"],
                        })
                        '''
                elif type == 2:
                    for index, result in enumerate(ddgs.news(
                        str(keywords),
                        **options
                    )):
                        results.append({
                            "title": result["title"],
                            "brief": result["body"],
                            "url": result["url"],
                            "source": result["source"],
                            "date": result["date"],
                        })
                        '''
                        results["origin"].append(result)
                        results["for_agent"].append({
                            "index": index,
                            "title": result["title"],
                            "body": result["body"],
                            "date": result["date"],
                        })
                        '''
        except Exception as e:
            results = f"No Result Because: { str(e) }"
        return results

    def search_definition(self, item_name: str, *, options: dict={}, proxy: str=None):
        return self.search(f'whatis { item_name }', options=options, proxy=proxy)

    def search_news(self, keywords: str, *, options: dict={}, proxy: str=None, timelimit: str=None):
        if timelimit:
            options.update({ "timelimit": timelimit })
        if "timelimit" not in options:
            options.update({ "timelimit": "w" })
        if "max_results" not in options:
            options.update({ "max_results": 10 })
        return self.search(keywords, options=options, proxy=proxy, type=2)

    def browse(self, url: str, *, retry_times:int = 0, proxy: str=None):
        ns_config = NsConfig()
        if proxy:
            if proxy.startswith("http:"):
                ns_config.proxies = { "http": proxy }
            elif proxy.startswith("https:"):
                ns_config.proxies = { "https": proxy }
        ns_config.request_timeout = 20
        ns_config.browser_user_agent = "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/119.0.0.0 Safari/537.36"
        ns_config.browser_accept = "*/*"
        ns_config.fetch_images = False
        ns_config.memoize_articles = False
        ns_article = NsArticle(url, config = ns_config)
        ns_article.download()
        ns_article.parse()
        result = ns_article.text
        return result
        '''
        if len(result) < 200:
            request_params = {}
            if proxy:
                request_params.update({ "proxies": proxy })
            response = httpx.get(url, **request_params)
            if response.status_code == 200:
                soup = BeautifulSoup(response.text, "html.parser")
                result = re.sub(r"\n{3,}", "\n\n", soup.get_text())
                return result
            else:
                if retry_times < 3:
                    return self.browse(url, retry_times = retry_times + 1)
                else:
                    return "Can not open the web page."
        else:
            return result
        '''

    def export(self):
        return {
            "search": {
                "desc": "search {keywords}.",
                "args": {
                    "keywords": ("String", "[*Required]keywords to search, seperate keywords by ' '."),
                    "options": {
                        "timelimit": ("String | Null", "'d': last day; 'w': this week; 'm': this month;")
                    },
                },
                "func": self.search,
                "require_proxy": True,
            },
            "search_info": {
                "desc": "search information about {keywords}.",
                "args": {
                    "keywords": ("String", "[*Required]keywords to search, seperate keywords by ' '."),
                    "options": {
                        "timelimit": ("String | Null", "'d': last day; 'w': this week; 'm': this month;")
                    },
                },
                "func": self.search,
                "require_proxy": True,
            },
            "search_news": {
                "desc": "search news about {keywords}.",
                "args": {
                    "keywords": ("String", "[*Required]keywords to search, seperate keywords by ' '."),
                    "options": {
                        "timelimit": ("String | Null", "'d': last day; 'w': this week; 'm': this month;")
                    },
                },
                "func": self.search_news,
                "require_proxy": True,
            },
            "search_definition": {
                "desc": "search definition about {item_name}.",
                "args": {
                    "item_name": ("String", "[*Required]item name to find the definition."),
                },
                "func": self.search_definition,
                "require_proxy": True,
            },
            "browse": {
                "desc": "Browse web page by url.",
                "args": {
                    "url": ("String", "[*Required]url to browse.")
                },
                "func": self.browse,
                "require_proxy": True,
            }
        }

def export():
    return ("Web", Web)


import json
import re
import yaml

def to_prompt_structure(prompt_dict: dict, layer_count: int=1, end: str=""):
    prompt = ""
    for key, content in prompt_dict.items():
        prompt += f"{ '#' * layer_count } { key }:\n"
        if isinstance(content, dict):
            prompt += to_prompt_structure(content, layer_count + 1) + '\n'
        else:
            prompt += str(content) + '\n'
        if layer_count == 1:
            prompt += "\n"
    if layer_count == 1:
        prompt += end
    return prompt

def to_instruction(origin):
    if origin == None:
        return None
    elif isinstance(origin, (list, tuple, set, dict)):
        return yaml.dump(origin, allow_unicode=True, sort_keys=False)
    else:
        return str(origin)

def to_json_desc(origin, layer_count = 0):
    if isinstance(origin, dict):
        json_string = ""
        if layer_count > 0:
            json_string += "\n"
        json_string += ("\t" * layer_count) + "{\n"
        for key, value in origin.items():
            json_string += ("\t" * (layer_count + 1)) + "\"" + key + "\": " + to_json_desc(value, layer_count + 1) + "\n"
        if layer_count > 0:
            json_string += ("\t" * (layer_count + 1)) + "},"
        else:
            json_string += "}"
        return json_string
    elif isinstance(origin, (list, set)):
        json_string = ""
        if layer_count > 0:
            json_string += "\n"
        json_string += ("\t" * layer_count) + "[\n"
        for item in origin:
            json_string += ("\t" * (layer_count + 1)) + to_json_desc(item, layer_count + 1) + ",\n"
        json_string += ("\t" * (layer_count + 1)) + "...\n"
        if layer_count > 0:
            json_string += ("\t" * layer_count) + "],"
        else:
            json_string += "]"
        return json_string
    elif isinstance(origin, tuple):
        if isinstance(origin[0], (dict, list, set)):
            json_string = f"\n{ to_json_desc(origin[0], layer_count + 1) },"
        else:
            json_string = f"<{ str(origin[0]) }>,"
        if len(origin) >= 2:
            json_string += f"//{ str(origin[1]) }"
        return json_string
    else:
        return str(origin)

def find_all_jsons(origin: str):
    pattern = r'"""(.*?)"""'
    origin = re.sub(
        pattern,
        lambda match: json.dumps(match.group(1)),
        origin,
        flags=re.DOTALL
    )
    origin = origin.replace("\"\"\"", "\"").replace("[OUTPUT]", "$<<OUTPUT>>")
    stage = 1
    json_blocks = []
    block_num = 0
    layer = 0
    skip_next = False
    in_quote = False
    for index, char in enumerate(origin):
        if skip_next:
            skip_next = False
            continue
        if stage == 1:
            if char == "\\":
                skip_next = True
                continue
            if char == "[" or char == "{":
                json_blocks.append(char)
                stage = 2
                layer += 1
                continue
        elif stage == 2:
            if not in_quote:
                if char == "\\":
                    skip_next = True
                    if origin[index + 1] == "\"":
                        char = "\""
                    else:
                        continue
                if char == "\"":
                    in_quote = True
                if char == "[" or char == "{":
                    layer += 1
                elif char == "]" or char == "}":
                    layer -= 1
                elif char in ("\t", " ", "\n"):
                    char = ""
                json_blocks[block_num] += char
            else:
                if char == "\\":
                    char += origin[index + 1]
                    skip_next = True
                elif char == "\n":
                    char = "\\n"
                elif char == "\t":
                    char = "\\t"
                elif char == "\"":
                    in_quote = not in_quote
                json_blocks[block_num] += char
            if layer == 0:
                json_blocks[block_num] = json_blocks[block_num].replace("$<<OUTPUT>>", "[OUTPUT]")
                block_num += 1
                stage = 1
    return json_blocks

def find_json(origin: str):
    result = find_all_jsons(origin)
    if len(result) > 0:
        return result[0]
    else:
        return None